@@ -15,20 +15,6 @@
  *	ACTIONS must be combined if 'actions together' is given.
  *	ACTIONS must be split if 'actions piecemeal' is given.
  *	ACTIONS must have current sources omitted for 'actions updated'.
- *
- * The CMD datatype holds a single command that is to be executed 
- * against a target, and they can chain together to represent the 
- * full collection of commands used to update a target.
- *
- * Structures:
- *
- * 	CMD - an action, ready to be formatted into a buffer and executed
- *
- * External routines:
- *
- * 	cmd_new() - return a new CMD or 0 if too many args
- *	cmd_free() - delete CMD and its parts
- *	cmd_next() - walk the CMD chain
  */
 
 /*
@@ -43,17 +29,12 @@ struct _cmd
 	CMD	*tail;		/* valid on in head */
 	RULE	*rule;		/* rule->actions contains shell script */
 	LIST	*shell;		/* $(SHELL) value */
+	int	exportvars;	/* nonzero to export variables using setenv */
 	LOL	args;		/* LISTs for $(<), $(>) */
-	char	*buf;		/* actual commands */
+	char	buf[ CMDBUF ];	/* actual commands */
 } ;
 
-CMD *cmd_new(
-	RULE	*rule,		/* rule (referenced) */
-	LIST	*targets,	/* $(<) (freed) */
-	LIST	*sources,	/* $(>) (freed) */
-	LIST	*shell,		/* $(SHELL) (freed) */
-	int	maxline );	/* max line length */
-
-void cmd_free( CMD *cmd );
+CMD 	*cmd_new();
+void	cmd_free();
 
 # define cmd_next( c ) ((c)->next)
