@@ -4,13 +4,46 @@
  * This file is part of Jam - see jam.c for Copyright information.
  */
 
+# include "jam.h"
+# include "execcmd.h"
+# include "lists.h"
+# include "variable.h"
+# include <errno.h>
+
+# if defined( unix ) || defined( NT ) || defined( __OS2__ )
+
+# if defined( _AIX) || \
+	(defined (COHERENT) && defined (_I386)) || \
+	defined(__sgi) || \
+	defined(__Lynx__) || \
+	defined(M_XENIX) || \
+	defined(__QNX__) || \
+	defined(__BEOS__) || \
+	defined(__ISC)
+# define vfork() fork()
+# endif
+
+# if defined( NT ) || defined( __OS2__ )
+
+# include <process.h>
+
+# if !defined( __BORLANDC__ ) && !defined( __OS2__ )
+# define wait my_wait
+static int my_wait(int *status);
+# endif
+
+# endif
+
+# include <unistd.h>
+# include <sys/wait.h>
+
 /*
- * execunix.c - execute a shell script on UNIX/WinNT/OS2/AmigaOS
+ * execunix.c - execute a shell script on UNIX/WinNT/OS2
  *
  * If $(JAMSHELL) is defined, uses that to formulate execvp()/spawnvp().
  * The default is:
  *
- *	/bin/sh -c %		[ on UNIX/AmigaOS ]
+ *	/bin/sh -c %		[ on UNIX ]
  *	cmd.exe /c %		[ on OS2/WinNT ]
  *
  * Each word must be an individual element in a jam variable value.
@@ -33,54 +66,27 @@
  * 05/04/94 (seiwald) - async multiprocess interface
  * 01/22/95 (seiwald) - $(JAMSHELL) support
  * 06/02/97 (gsar)    - full async multiprocess support for Win32
- * 01/20/00 (seiwald) - Upgraded from K&R to ANSI C
- * 11/04/02 (seiwald) - const-ing for string literals
- * 12/27/02 (seiwald) - grist .bat file with pid for system uniqueness
- * 05/06/05 (seiwald) - new execmax() to return max command line len.
  */
 
-# include "jam.h"
-# include "lists.h"
-# include "execcmd.h"
-# include <errno.h>
-
-# ifdef USE_EXECUNIX
-
-# ifdef OS_OS2
-# define USE_EXECNT
-# include <process.h>
-# endif
-
-# ifdef OS_MACOSX
-# include <unistd.h>
-# endif
-
-# ifdef OS_NT 
-# define USE_EXECNT
-# include <process.h>
-# define WIN32_LEAN_AND_MEAN
-# include <windows.h>		/* do the ugly deed */
-# define USE_MYWAIT
-# if !defined( __BORLANDC__ ) 
-# define wait my_wait
-static long my_wait( int *status );
-# endif
-# endif
-
 static int intr = 0;
+
 static int cmdsrunning = 0;
+
+# ifdef NT
 static void (*istat)( int );
+void onintr( int );
+# else
+static void (*istat)();
+#endif
 
 static struct
 {
-	long	pid; /* on win32, a real process handle */
-	void	(*func)( void *closure, int status );
+	int	pid;		/* on win32, a real process handle */
+	void	(*func)();
 	void 	*closure;
-
-# ifdef USE_EXECNT
+# if defined( NT ) || defined( __OS2__ )
 	char	*tempfile;
 # endif
-
 } cmdtab[ MAXJOBS ] = {{0}};
 
 /*
@@ -88,87 +94,65 @@ static struct
  */
 
 void
-onintr( int disp )
+onintr( disp )
+int disp;
 {
 	intr++;
+#ifdef APPLE_EXTENSIONS
+	pbx_printf( "JAM ", "...interrupted\n" );
+#else
 	printf( "...interrupted\n" );
+#endif
 }
 
-/*
- * execmax() - max permitted string to execcmd()
- */
-
-# ifdef OS_NT
-
-int
-execmax()
-{
-	static int maxline = 0;
-	OSVERSIONINFO o;
-
-	if( maxline )
-	    return maxline;
-
-	/* Get from OS */
-
-	/* Beyond Win 2003?  (6+?) -- 8191 */
-	/* WinXP (5/0) or Win 2003 (5/1) -- 8191 */
-	/* Win 2k (5/0) or Win NT 4 (4/0) -- 2047 */
-	/* older -- 996 */
-
-	o.dwOSVersionInfoSize = sizeof( o );
-
-	if( !GetVersionEx( &o ) )
-	    maxline = MAXLINE;
-	else if( o.dwMajorVersion >= 6 )
-	    maxline = 8191;
-	else if( o.dwMajorVersion == 5 && o.dwMinorVersion >= 1 )
-	    maxline = 8191;
-	else if( o.dwMajorVersion == 5 && o.dwMinorVersion == 0 )
-	    maxline = 2047;
-	else if( o.dwMajorVersion == 4 && o.dwMinorVersion == 0 )
-	    maxline = 2047;
-	else							  
-	    maxline = MAXLINE;
-
-	if( DEBUG_EXECCMD )
-	    printf( "Windows %d/%d maxline %d\n",
-	    	o.dwMajorVersion,
-		o.dwMinorVersion,
-		maxline );
-
-	return maxline;
-}
-
-# else
-
-int
-execmax()
-{
-	return MAXLINE;
-}
-
-# endif
-
 /*
  * execcmd() - launch an async command execution
  */
 
 void
-execcmd( 
-	char *string,
-	void (*func)( void *closure, int status ),
-	void *closure,
-	LIST *shell )
+execcmd( string, func, closure, shell, exportvars )
+char *string;
+void (*func)();
+void *closure;
+LIST *shell;
+int exportvars;
 {
 	int pid;
 	int slot;
-	const char *argv[ MAXARGC + 1 ];	/* +1 for NULL */
+	char *argv[ MAXARGC + 1 ];	/* +1 for NULL */
+#ifdef APPLE_EXTENSIONS
+	// :mferris:20001213 We use a pipe to send commands to sh through stdin instead of passing them as an argument with "-c" since there are problems with the command line arguments somehow getting mangled when they contain weird UTF-8 encoded file names...
+	int pipeForStdInToShell[2];
+	static char usePipeForShellCommands = -1;
+#endif
 
-# ifdef USE_EXECNT
+# if defined( NT ) || defined( __OS2__ )
+	static char *comspec;
 	char *p;
+
+	/* XXX this is questionable practice, since COMSPEC has
+	 * a high degree of variability, resulting in Jamfiles
+	 * that frequently won't work.  COMSPEC also denotes a shell
+	 * fit for interative use, not necessarily/merely a shell
+	 * capable of launching commands.  Besides, people can
+	 * just set JAMSHELL instead.
+	 */
+	if( !comspec && !( comspec = getenv( "COMSPEC" ) ) )
+	    comspec = "cmd.exe";
 # endif
 
+#ifdef APPLE_EXTENSIONS
+	if( usePipeForShellCommands == -1 ) {
+	    if( getenv( "USE_CMDLINE_FOR_SHELL_COMMANDS" ) != NULL ) {
+		/*printf("NOT USING PIPES\n");*/
+		usePipeForShellCommands = 0;
+	    } else {
+		/*printf("USING PIPES\n");*/
+		usePipeForShellCommands = 1;
+	    }
+	}
+#endif
+
 	/* Find a slot in the running commands table for this one. */
 
 	for( slot = 0; slot < MAXJOBS; slot++ )
@@ -181,7 +165,7 @@ execcmd(
 	    exit( EXITBAD );
 	}
 
-# ifdef USE_EXECNT
+# if defined( NT ) || defined( __OS2__ )
 	if( !cmdtab[ slot ].tempfile )
 	{
 	    char *tempdir;
@@ -190,12 +174,10 @@ execcmd(
 		!( tempdir = getenv( "TMP" ) ) )
 		    tempdir = "\\temp";
 
-	    /* +32 is room for \jamXXXXXtSS.bat (at least) */
+	    cmdtab[ slot ].tempfile = malloc( strlen( tempdir ) + 14 );
 
-	    cmdtab[ slot ].tempfile = malloc( strlen( tempdir ) + 32 );
-
-	    sprintf( cmdtab[ slot ].tempfile, "%s\\jam%dt%d.bat", 
-				tempdir, GetCurrentProcessId(), slot );
+	    sprintf( cmdtab[ slot ].tempfile, "%s\\jamtmp%02d.bat", 
+				tempdir, slot );
 	}
 
 	/* Trim leading, ending white space */
@@ -208,12 +190,12 @@ execcmd(
 	while( p && isspace( *p ) )
 		++p;
 
-	/* If multi line, or too long, or JAMSHELL is set, write to bat file. */
+	/* If multi line or too long, write to bat file. */
 	/* Otherwise, exec directly. */
 	/* Frankly, if it is a single long line I don't think the */
 	/* command interpreter will do any better -- it will fail. */
 
-	if( p && *p || strlen( string ) > MAXLINE || shell )
+	if( p && *p || strlen( string ) > MAXLINE )
 	{
 	    FILE *f;
 
@@ -226,10 +208,10 @@ execcmd(
 	    string = cmdtab[ slot ].tempfile;
 	}
 # endif
-
+	
 	/* Forumulate argv */
 	/* If shell was defined, be prepared for % and ! subs. */
-	/* Otherwise, use stock /bin/sh (on unix) or cmd.exe (on NT). */
+	/* Otherwise, use stock /bin/sh (on unix) or comspec (on NT). */
 
 	if( shell )
 	{
@@ -258,15 +240,29 @@ execcmd(
 	}
 	else
 	{
-# ifdef USE_EXECNT
-	    argv[0] = "cmd.exe";
+# if defined( NT ) || defined( __OS2__ )
+	    argv[0] = comspec;
 	    argv[1] = "/Q/C";		/* anything more is non-portable */
-# else
-	    argv[0] = "/bin/sh";
-	    argv[1] = "-c";
-# endif
 	    argv[2] = string;
 	    argv[3] = 0;
+# else /* ! ( defined( NT ) || defined( __OS2__ ) ) */
+	    argv[0] = "/bin/sh";
+#ifdef APPLE_EXTENSIONS
+	    if (usePipeForShellCommands) {
+		argv[1] = "-s";
+		argv[2] = NULL;
+	    } else {
+		argv[1] = "-c";
+		argv[2] = string;
+		argv[3] = 0;
+	    }
+#else /* ! ( APPLE_EXTENSIONS ) */
+	    argv[1] = "-c";
+	    argv[2] = string;
+	    argv[3] = 0;
+#endif /* ! ( APPLE_EXTENSIONS ) */
+
+# endif /* ! ( defined( NT ) || defined( __OS2__ ) ) */
 	}
 
 	/* Catch interrupts whenever commands are running. */
@@ -274,25 +270,54 @@ execcmd(
 	if( !cmdsrunning++ )
 	    istat = signal( SIGINT, onintr );
 
+#ifdef APPLE_EXTENSIONS
+	/* Set up pipes, etc so that we can correctly multiplex output from multiple simultaneous commands */
+    #if 0 // not yet implemented -- we assume one command at a time for now
+	{
+	    int   stdout_pipe[2], stderr_pipe[2];
+
+	    pipe(stdout_pipe);
+	    pipe(stderr_pipe);
+	    printf("stdout_pipe[0]=%i, stdout_pipe[1]=%i\n", stdout_pipe[0], stdout_pipe[1]);
+	    printf("stderr_pipe[0]=%i, stderr_pipe[1]=%i\n", stderr_pipe[0], stderr_pipe[1]);
+	}
+    #endif
+#endif
+
 	/* Start the command */
 
-# ifdef USE_EXECNT
-	if( ( pid = spawnvp( P_NOWAIT, argv[0], argv ) ) == -1 )
+# if defined( NT ) || defined( __OS2__ )
+	if( ( pid = spawnvp( P_NOWAIT, argv[0], argv ) ) < 0 )
 	{
 	    perror( "spawn" );
 	    exit( EXITBAD );
 	}
 # else
-# ifdef NO_VFORK
-	if ((pid = fork()) == 0) 
-# else
+#ifdef APPLE_EXTENSIONS
+	if (usePipeForShellCommands) {
+	    // Create the pipe we will use to send stdin to child
+	    pipe(pipeForStdInToShell);
+	}
+#endif
+
 	if ((pid = vfork()) == 0) 
-# endif
    	{
-	    /* hpux doesn't like const here */
-	    execvp( argv[0], (char **)argv );
+#ifdef APPLE_EXTENSIONS
+	    if (usePipeForShellCommands) {
+		// CHILD: Close write end of pipe in child and make stdin be the read end.
+		close(pipeForStdInToShell[1]);
+		dup2(pipeForStdInToShell[0], 0);
+	    }
+#endif
+	    if (exportvars)
+		var_setenv_all_exported_variables();
+	    //else
+		//printf("NOT ");
+	    //printf("doing setenv for exec of %s\n", string);
+	    execvp( argv[0], argv );
 	    _exit(127);
 	}
+
 	if( pid == -1 )
 	{
 	    perror( "vfork" );
@@ -305,6 +330,24 @@ execcmd(
 	cmdtab[ slot ].func = func;
 	cmdtab[ slot ].closure = closure;
 
+#ifdef APPLE_EXTENSIONS
+	if (usePipeForShellCommands) {
+	    // PARENT: 
+	    // Close read end of pipe in parent.
+	    close(pipeForStdInToShell[0]);
+	    // Write the script string to the write end of the pipe, then close that one too.
+	    {
+		int toWrite = strlen(string);
+		int written = write(pipeForStdInToShell[1], string, toWrite);
+		
+		if (toWrite != written) {
+		    printf("Error: only wrote %d bytes of %d bytes of command to sub-shell.", written, toWrite);
+		}
+	    }
+	    close(pipeForStdInToShell[1]);
+	}
+#endif
+
 	/* Wait until we're under the limit of concurrent commands. */
 	/* Don't trust globs.jobs alone. */
 
@@ -353,12 +396,6 @@ execwait()
 	    exit( EXITBAD );
 	}
 
-# ifdef USE_EXECNT
-	/* Clear the temp file */
-
-	unlink( cmdtab[ i ].tempfile );
-# endif
-
 	/* Drive the completion */
 
 	if( !--cmdsrunning )
@@ -372,16 +409,21 @@ execwait()
 	    rstat = EXEC_CMD_OK;
 
 	cmdtab[ i ].pid = 0;
-
+	
 	(*cmdtab[ i ].func)( cmdtab[ i ].closure, rstat );
 
 	return 1;
 }
 
-# ifdef USE_MYWAIT
+# if defined( NT ) && !defined( __BORLANDC__ )
 
-static long
-my_wait( int *status )
+# define WIN32_LEAN_AND_MEAN
+
+# include <windows.h>		/* do the ugly deed */
+
+static int
+my_wait( status )
+int *status;
 {
 	int i, num_active = 0;
 	DWORD exitcode, waitcode;
@@ -427,7 +469,7 @@ my_wait( int *status )
 	    if ( GetExitCodeProcess(active_handles[i], &exitcode) ) {
 		CloseHandle(active_handles[i]);
 		*status = (int)((exitcode & 0xff) << 8);
-		return (long)active_handles[i];
+		return (int)active_handles[i];
 	    }
 	}
 
@@ -437,6 +479,6 @@ FAILED:
     
 }
 
-# endif /* USE_MYWAIT */
+# endif /* NT && !__BORLANDC__ */
 
-# endif /* USE_EXECUNIX */
+# endif /* unix || NT || __OS2__ */
