@@ -1,6 +1,6 @@
 /*
  * /+\
- * +\	Copyright 1993-2002 Christopher Seiwald and Perforce Software, Inc.
+ * +\	Copyright 1993, 1996 Christopher Seiwald.
  * \+/
  *
  * This file is part of jam.
@@ -12,10 +12,42 @@
  * ALL WARRANTIES ARE HEREBY DISCLAIMED.
  */
 
+# include "jam.h"
+# include "option.h"
+# include "make.h"
+# ifdef FATFS
+# include "patchlev.h"
+# else
+# include "patchlevel.h"
+# endif
+
+/* These get various function declarations. */
+
+# include "lists.h"
+# include "parse.h"
+# include "variable.h"
+# include "compile.h"
+# include "rules.h"
+# include "newstr.h"
+# include "scan.h"
+# ifdef FATFS
+# include "timestam.h"
+# else
+# include "timestamp.h"
+# endif
+
+/* Macintosh is "special" */
+
+# ifdef macintosh
+# include <QuickDraw.h>
+# else
+# include <stdarg.h>
+# endif
+
 /*
  * jam.c - make redux
  *
- * See Jam.html for usage information.
+ * See jam(1) and Jamfile(5) for usage information.
  *
  * These comments document the code.
  *
@@ -30,18 +62,14 @@
  *       /   |    \              \
  *      /    |     \             |
  *  scan     |     compile      make
- *   |       |    /  | \       / |  \
- *   |       |   /   |  \     /  |   \
- *   |       |  /    |   \   /   |    \
- * jambase parse     |   rules  search make1
- *                   |           |      |   \
- *                   |           |      |    \
- *                   |           |      |     \
- *               builtins    timestamp command execute
- *                               |
- *                               |
- *                               |
- *                             filesys
+ *   |       |    /    \       / |  \
+ *   |       |   /      \     /  |   \
+ *   |       |  /        \   /   |    \
+ * jambase parse         rules  search make1
+ *                               |	|   \
+ *                               |	|    \
+ *                               |	|     \
+ *                           timestamp command execute
  *
  *
  * The support routines are called by all of the above, but themselves
@@ -51,7 +79,7 @@
  *                      /  |   |   |
  *                     /   |   |   |
  *                    /    |   |   |
- *                 lists   |   |   pathsys
+ *                 lists   |   |   filesys
  *                    \    |   |
  *                     \   |   |
  *                      \  |   |
@@ -63,13 +91,13 @@
  *
  * Roughly, the modules are:
  *
- *	builtins.c - jam's built-in rules
  *	command.c - maintain lists of commands
  *	compile.c - compile parsed jam statements
  *	execunix.c - execute a shell script on UNIX
  *	execvms.c - execute a shell script, ala VMS
  *	expand.c - expand a buffer, given variable values
- *	file*.c - scan directories and archives on *
+ *	fileunix.c - manipulate file names and scan directories on UNIX
+ *	filevms.c - manipulate file names and scan directories on VMS
  *	hash.c - simple in-memory hashing routines 
  *	headers.c - handle #includes in source files
  *	jambase.c - compilable copy of Jambase
@@ -80,8 +108,6 @@
  *	newstr.c - string manipulation routines
  *	option.c - command line option processing
  *	parse.c - make and destroy parse trees as driven by the parser
- *	path*.c - manipulate file names on *
- *	hash.c - simple in-memory hashing routines 
  *	regexp.c - Henry Spencer's regexp
  *	rules.c - access to RULEs, TARGETs, and ACTIONs
  *	scan.c - the jam yacc scanner
@@ -92,108 +118,142 @@
  * 05/04/94 (seiwald) - async multiprocess (-j) support
  * 02/08/95 (seiwald) - -n implies -d2.
  * 02/22/95 (seiwald) - -v for version info.
- * 09/11/00 (seiwald) - PATCHLEVEL folded into VERSION.
- * 01/10/01 (seiwald) - pathsys.h split from filesys.h
- * 01/21/02 (seiwald) - new -q to quit quickly on build failure
- * 03/16/02 (seiwald) - support for -g (reorder builds by source time)
- * 09/19/02 (seiwald) - new -d displays
- * 10/22/02 (seiwald) - list_new() now does its own newstr()/copystr()
- * 11/04/02 (seiwald) - const-ing for string literals
  */
 
-# include "jam.h"
-# include "option.h"
-# include "patchlevel.h"
-
-/* These get various function declarations. */
-
-# include "lists.h"
-# include "parse.h"
-# include "variable.h"
-# include "compile.h"
-# include "builtins.h"
-# include "rules.h"
-# include "newstr.h"
-# include "scan.h"
-# include "timestamp.h"
-# include "make.h"
-
-/* Macintosh is "special" */
-
-# ifdef OS_MAC
-# include <QuickDraw.h>
-# endif
-
-/* And UNIX for this */
-
-# ifdef unix
-# include <sys/utsname.h>
-# endif
-
 struct globs globs = {
 	0,			/* noexec */
 	1,			/* jobs */
-	0,			/* quitquick */
-	0,			/* newestfirst */
-# ifdef OS_MAC
-	{ 0 },			/* display - suppress actions output */
+	0,			/* ignore */
+#ifdef APPLE_EXTENSIONS
+	0,                      /* apple_jam_extensions */
+	0,			/* parsable_output */
+	NULL,			/* cmdline_defines */
+#endif
+# ifdef macintosh
+	{ 0, 0 }		/* debug - suppress tracing output */
 # else
-	{ 0, 1 }, 		/* display actions  */
+	{ 0, 1 } 		/* debug ... */
 # endif
-	0			/* output commands, not run them */
 } ;
 
+
+#ifdef APPLE_EXTENSIONS
+
+int pbx_printf( const char * category_tag, const char * format, ... )
+{
+    va_list   args;
+    int       result = 0;
+
+    va_start(args, format);
+    if (category_tag != NULL)
+	printf("[%s]", category_tag);
+    result = vprintf(format, args);
+    va_end(args);
+    return result;
+}
+
+#endif
+
+
 /* Symbols to be defined as true for use in Jambase */
 
-static const char *othersyms[] = { OSMAJOR, OSMINOR, OSPLAT, JAMVERSYM, 0 } ;
+static char *othersyms[] = { OSSYMS OSPLATSYM, JAMVERSYM, 0 } ;
 
-/* Known for sure: 
- *	mac needs arg_enviro
- *	OS2 needs extern environ
- */
-
-# ifdef OS_MAC
-# define use_environ arg_environ
-# ifdef MPW
-QDGlobals qd;
-# endif
-# endif
-
-# ifndef use_environ
-# define use_environ environ
-# if !defined( __WATCOM__ ) && !defined( OS_OS2 ) && !defined( OS_NT ) 
+# ifndef __WATCOM__
+# ifndef __OS2__
+# ifndef NT
 extern char **environ;
 # endif
 # endif
+# endif
 
-main( int argc, char **argv, char **arg_environ )
+# ifdef macintosh
+char** demakeifyArguments( int argc, char **argv )
+# else
+char** demakeifyArguments( argc, argv )
+char    **argv;
+# endif
 {
+	char**	newArgvBuffer = (char**)calloc(sizeof(char*), argc);
+	int	i;
+	int	lastOption = 0;
+	int	j;
+
+	for (i=0; i<argc; i++)
+	{
+#ifdef DEBUG_JAM_INPUT_ARGUMENTS
+		printf("argv[%2u]='%s'\n", i, argv[i]);
+#endif
+		newArgvBuffer[i] = NULL;
+	}
+
+	for (i=0; i<argc; i++) {
+		if (argv[i][0] == '-')
+			lastOption = i+1;
+		else if (strchr (argv[i], '=')) {
+			char* old = argv[i];
+			char* new = malloc(strlen(old)+3);
+			sprintf(new, "-s%s", old);
+			for (j=i; j>lastOption; j--)
+				argv[j] = argv[j-1];
+			newArgvBuffer[lastOption] = old;
+			argv[lastOption++] = new;
+		}
+	}
+	return newArgvBuffer;
+}
+
+# ifdef macintosh
+void remakeifyArguments( int argc, char **argv , char** newArgvBuffer )
+# else
+void remakeifyArguments( argc, argv, newArgvBuffer  )
+char    **argv;
+char    **newArgvBuffer;
+# endif
+{
+	int	i;
+
+	for (i=0; i<argc; i++) if (newArgvBuffer[i] != NULL) {
+		free (argv[i]);
+		argv[i] = newArgvBuffer[i];
+	}
+	free (newArgvBuffer);
+}
+
+# ifdef macintosh
+int main( int argc, char **argv, char **environ )
+# else
+int main( argc, argv )
+char	**argv;
+# endif
+{
+	const unsigned  realArgc = argc;
+	char		**newArgvBuffer;
 	int		n;
-	const char	*s;
+	char		*s;
 	struct option	optv[N_OPTS];
-	const char	*all = "all";
+	char		*all = "all";
 	int		anyhow = 0;
 	int		status;
 
-# ifdef OS_MAC
+# ifdef macintosh
 	InitGraf(&qd.thePort);
 # endif
 
 	argc--, argv++;
 
-	if( ( n = getoptions( argc, argv, "d:j:f:gs:t:ano:qv", optv ) ) < 0 )
+	newArgvBuffer = demakeifyArguments(argc, argv);
+
+	if( ( n = getoptions( argc, argv, "d:j:f:s:t:aknv", optv ) ) < 0 )
 	{
 	    printf( "\nusage: jam [ options ] targets...\n\n" );
 
             printf( "-a      Build all targets, even if they are current.\n" );
-            printf( "-dx     Display (a)actions (c)causes (d)dependencies\n" );
-	    printf( "        (m)make tree (x)commands (0-9) debug levels.\n" );
+            printf( "-dx     Set the debug level to x (0-9).\n" );
             printf( "-fx     Read x instead of Jambase.\n" );
-	    printf( "-g      Build from newest sources first.\n" );
             printf( "-jx     Run up to x shell commands concurrently.\n" );
+            printf( "-k      Ignore build errors.\n" );
             printf( "-n      Don't actually execute the updating actions.\n" );
-            printf( "-ox     Write the updating actions to file x.\n" );
-            printf( "-q      Quit quickly as soon as a target fails.\n" );
 	    printf( "-sx=y   Set variable x=y, overriding environment.\n" );
             printf( "-tx     Rebuild x, even if it is up-to-date.\n" );
             printf( "-v      Print the version of jam and exit.\n\n" );
@@ -207,20 +267,22 @@ main( int argc, char **argv, char **arg_environ )
 
 	if( ( s = getoptval( optv, 'v', 0 ) ) )
 	{
-	    printf( "Jam %s. %s. ", VERSION, OSMINOR );
-	    if( *OSPLAT ) printf( "%s. ", OSPLAT );
-	    printf( "Copyright 1993-2010 Christopher Seiwald.\n" );
-
+	    printf( "Jam/MR  " );
+	    printf( "Version %s.%s.  ", VERSION, PATCHLEVEL );
+	    printf( "Copyright 1993, 1997 Christopher Seiwald.\n" );
+#ifdef APPLE_EXTENSIONS
+	    printf( "with Apple ProjectBuilder Extensions\n" );
+#endif
 	    return EXITOK;
 	}
 
 	/* Pick up interesting options */
 
-	if( ( s = getoptval( optv, 'n', 0 ) ) )
-	    globs.noexec++, DEBUG_MAKE = DEBUG_MAKEQ = DEBUG_EXEC = 1; 
+    if( ( s = getoptval( optv, 'n', 0 ) ) )
+        globs.noexec++, globs.debug[2] = 1;
 
-	if( ( s = getoptval( optv, 'q', 0 ) ) )
-	    globs.quitquick = 1;
+    if( ( s = getoptval( optv, 'k', 0 ) ) )
+        globs.ignore++;
 
 	if( ( s = getoptval( optv, 'a', 0 ) ) )
 	    anyhow++;
@@ -228,110 +290,87 @@ main( int argc, char **argv, char **arg_environ )
 	if( ( s = getoptval( optv, 'j', 0 ) ) )
 	    globs.jobs = atoi( s );
 
-	if( ( s = getoptval( optv, 'g', 0 ) ) )
-	    globs.newestfirst = 1;
+#ifdef APPLE_EXTENSIONS
+	if( getenv( "NATIVE_ARCH" ) != NULL ) {
+	    globs.apple_jam_extensions = 1;
+	}
+	if( getenv( "ENABLE_APPLE_JAM_OUTPUT_ANNOTATION" ) != NULL ) {
+	    globs.parsable_output = 1;
+	}
+#endif
 
 	/* Turn on/off debugging */
 
-	for( n = 0; s = getoptval( optv, 'd', n ); n++ )
+	for( n = 0; (s = getoptval( optv, 'd', n )) != NULL; n++ )
 	{
-	    int i = atoi( s );
+	    int i;
 
 	    /* First -d, turn off defaults. */
 
 	    if( !n )
-		DEBUG_MAKE = DEBUG_MAKEQ = DEBUG_EXEC = 0;
+		for( i = 0; i < DEBUG_MAX; i++ )
+		    globs.debug[i] = 0;
 
-	    /* n turns on levels 1-n */
-	    /* +n turns on level n */
-	    /* c turns on named display c */
+	    i = atoi( s );
 
 	    if( i < 0 || i >= DEBUG_MAX )
 	    {
 		printf( "Invalid debug level '%s'.\n", s );
+		continue;
 	    }
-	    else if( *s == '+' )
-	    {
+
+	    /* n turns on levels 1-n */
+	    /* +n turns on level n */
+
+	    if( *s == '+' )
 		globs.debug[i] = 1;
-	    }
-	    else if( i ) while( i )
-	    {
+	    else while( i )
 		globs.debug[i--] = 1;
-	    }
-	    else while( *s ) switch( *s++ )
-	    {
-	    case 'a': DEBUG_MAKE = DEBUG_MAKEQ = 1; break;
-	    case 'c': DEBUG_CAUSES = 1; break;
-	    case 'd': DEBUG_DEPENDS = 1; break;
-	    case 'm': DEBUG_MAKEPROG = 1; break;
-	    case 'r': DEBUG_COMPILE = 1; break;
-	    case 'x': DEBUG_EXEC = 1; break;
-	    case '0': break;
-	    default: printf( "Invalid debug flag '%c'.\n", s[-1] );
-	    }
 	}
 
 	/* Set JAMDATE first */
 
 	{
-	    char buf[ 128 ];
+	    char *date;
 	    time_t clock;
 	    time( &clock );
-	    strcpy( buf, ctime( &clock ) );
+	    date = newstr( ctime( &clock ) );
 
 	    /* Trim newline from date */
 
-	    if( strlen( buf ) == 25 )
-		buf[ 24 ] = 0;
+	    if( strlen( date ) == 25 )
+		date[ 24 ] = 0;
 
-	    var_set( "JAMDATE", list_new( L0, buf, 0 ), VAR_SET );
+	    var_set( "JAMDATE", list_new( L0, newstr( date ) ), VAR_SET, 1 /* export-in-environment */ );
 	}
 
-	/* And JAMUNAME */
-# ifdef unix
-	{
-	    struct utsname u;
-
-	    if( uname( &u ) >= 0 )
-	    {
-		LIST *l = L0;
-		l = list_new( l, u.machine, 0 );
-		l = list_new( l, u.version, 0 );
-		l = list_new( l, u.release, 0 );
-		l = list_new( l, u.nodename, 0 );
-		l = list_new( l, u.sysname, 0 );
-		var_set( "JAMUNAME", l, VAR_SET );
-	    }
-	}
-# endif /* unix */
-
-	/*
-	 * Jam defined variables OS, OSPLAT
-	 */
-
-	var_defines( othersyms );
-
 	/* load up environment variables */
 
-	var_defines( (const char **)use_environ );
+	var_defines( environ, 0 /* don't export-in-environment */ );
+	var_defines( othersyms, 1 /* export-in-environment */ );
 
 	/* Load up variables set on command line. */
 
-	for( n = 0; s = getoptval( optv, 's', n ); n++ )
+	globs.cmdline_defines = (const char **)malloc(sizeof(char *) * (realArgc + 1));
+	for( n = 0; (s = getoptval( optv, 's', n )) != NULL; n++ )
 	{
-	    const char *symv[2];
+	    char *symv[2];
 	    symv[0] = s;
 	    symv[1] = 0;
-	    var_defines( symv );
+	    var_defines( symv, 1 /* export-in-environment */ );
+	  #ifdef APPLE_EXTENSIONS
+	    globs.cmdline_defines[n] = s;
+	  #endif
 	}
+	globs.cmdline_defines[n] = NULL;
 
-	/* Initialize built-in rules */
+	/* Initialize builtins */
 
-	load_builtins();
+	compile_builtins();
 
 	/* Parse ruleset */
 
-	for( n = 0; s = getoptval( optv, 'f', n ); n++ )
+	for( n = 0; (s = getoptval( optv, 'f', n )) != NULL; n++ )
 	    parse_file( s );
 
 	if( !n )
@@ -341,27 +380,15 @@ main( int argc, char **argv, char **arg_environ )
 
 	/* Manually touch -t targets */
 
-	for( n = 0; s = getoptval( optv, 't', n ); n++ )
+	for( n = 0; (s = getoptval( optv, 't', n )) != NULL; n++ )
 	    touchtarget( s );
 
-	/* If an output file is specified, set globs.cmdout to that */
-
-	if( s = getoptval( optv, 'o', 0 ) )
-	{
-	    if( !( globs.cmdout = fopen( s, "w" ) ) )
-	    {
-		printf( "Failed to write to '%s'\n", s );
-		exit( EXITBAD );
-	    }
-	    globs.noexec++;
-	}
-
 	/* Now make target */
 
 	if( !argc )
 	    status |= make( 1, &all, anyhow );
 	else
-	    status |= make( argc, (const char **)argv, anyhow );
+	    status |= make( argc, argv, anyhow );
 
 	/* Widely scattered cleanup */
 
@@ -370,10 +397,9 @@ main( int argc, char **argv, char **arg_environ )
 	donestamps();
 	donestr();
 
-	/* close cmdout */
-
-	if( globs.cmdout )
-	    fclose( globs.cmdout );
+#ifndef APPLE_EXTENSIONS
+	remakeifyArguments(argc, argv, newArgvBuffer);
+#endif
 
 	return status ? EXITBAD : EXITOK;
 }
